(* Title:  BaseDefinitions/problem-def.sml
   Author: Walther Neuper
   (c) due to copyright terms

Here is a minimum of code required for Know_Store.thy.
For main code see structure Problem.
*)
signature PROBLEM_DEFINITION =
sig
  type id = References_Def.id
  val id_empty: id
  val id_to_string: id -> string

  type T
  val empty: T

  type store
  val empty_store: T Store.node

  val check_unique: Check_Unique.id -> T Store.T -> unit
  val s_to_string: T list -> string
end

(**)
structure Probl_Def(**): PROBLEM_DEFINITION(**) =
struct
(**)

type id = References_Def.id;
val id_empty = References_Def.empty_probl_id;
val id_to_string = References_Def.id_to_string;

type T = 
  {guh : Check_Unique.id,    (* unique within Isac_Knowledge, for html-presentation      *)
  mathauthors : string list, (* copyright                                                *)
  init : References_Def.id,  (* to start refinement with                                 *)
  thy  : theory,      (* which allows to compile that T
                        TODO: search generalized for subthy (ref.p.69*)
                        (*^^^ WN050912 NOT used during application of the problem,
                        because applied terms may be from 'subthy' as well as from super;
                        thus we take 'maxthy'; see match_ags !                           *)
  cas : term option,         (* CAS_Cmd                                                  *)
  met : References_Def.id list,   (* methods solving the T                               *)
  prls : Rule_Set.T,         (* for evaluation of preconditions in "#Where" on "#Given"  *)
  where_ : term list,        (* ? DEL, as soon as they are input interactively ?         *)
  ppc : Model_Pattern.T      (* contains "#Given", "#Where", "#Find", "#Relate"
                                for constraints on identifiers see "O_Model.copy_name"   *)
}   
val empty = {guh = "pbl_empty", mathauthors = [], init = id_empty, thy = @{theory "Pure"},
  cas = NONE, prls = Rule_Set.Empty, where_ = [], ppc = [], met = []} : T
fun pbt2str ({cas = cas', guh = guh', init = init', mathauthors = ma', met = met', ppc = ppc',
      prls = prls', thy = thy', where_ = w'} : T)
    = "{cas = " ^ (UnparseC.term_opt cas') ^  ", guh = \"" ^ guh'  ^ "\", init = "
      ^ (strs2str init') ^ ", mathauthors = " ^ (strs2str ma' |> quote) ^ ", met = "
      ^ (strslist2strs met') ^ ", ppc = " ^ Model_Pattern.to_string ppc' ^ ", prls = "
      ^ (Rule_Set.id prls' |> quote) ^ ", thy = {" ^ Context.theory_name thy' ^ "}, where_ = "
      ^ (UnparseC.terms w') ^ "}" |> linefeed;
fun s_to_string pbts = map pbt2str pbts |> list2str;

val empty_store = Store.Node ("empty_probl_id", [empty], [])
type store = (T Store.node) list

val check_unique =
  Check_Unique.message (Check_Unique.collect (#guh : T -> Check_Unique.id));

(**)end(*struct*)
